unit SUtil;

interface

uses SUtilBase;

{ Type Convert Function }

function IStr(Value: Integer): string;
function FStr(Value: Double; Digits: Integer): string;
function IZStr(Value, Width: Integer): string;
function SInt(const S: string): Integer;
function SFloat(const S: string): Double;
function UnitFormat(Value: Comp): string;
function RoundOff(Value: Double): Integer;
function NumFormat(Value: Integer): string;

{ Strings Function }

function ConvLineBreak(const S: string; const LineBreak: string): string;
function CopyCStr(const Src: string; Dst: PChar; Size: Cardinal): Cardinal;

{ Strings Function (Locale depend) }

function Lower(const S: string): string;
function Upper(const S: string): string;
function LStr(const S: string; Count: Integer): string;
function RStr(const S: string; Count: Integer): string;
function RevStr(const S: string): string;
function StrColl(const S1, S2: string): Integer;

{ Path Function }

function IsFile(const Path: string): Boolean;
function IsDir(const Path: string): Boolean;
function AbsPath(const Path: string): string;
procedure SplitDrive(const Path: string; var Drive, Tail: string);
procedure SplitPath(const Path: string; var Head, Tail: string);
function BaseName(const Path: string): string;
function DirName(const Path: string): string;
procedure SplitExt(const Path: string; var Root, Ext: string);
function JoinPath(const Path1, Path2: string): string;
function ExpandVars(const Path: string): string;

{ Operatiing System Services }

function GetEnv(const Name: string; const Default: string): string;
function SysError(Code: Integer): string;

{ Exceptions }

type
  ESException = class(TObject)
    Msg: string;
    constructor Create(const AMsg: string);
  end;

  ESIOError = class(ESException)
    ErrNo: Integer;
    FileName: string;
    constructor Create(const AFileName: string);
  end;

  ESOSError = class(ESException)
    ErrNo: Integer;
    constructor Create;
  end;

{ Classes }

type
  TSFile = class(_TSFile)
  private
    FReadP, FEndP, FBuf: PChar;
  public
    destructor Destroy; override;
    function ReadLine(var S: string): Boolean;
  end;

  TListDir = class(_TListDir);

implementation

{$I SUtil.inc}

{ Type Convert Function }

function IStr(Value: Integer): string;
begin
  Str(Value, Result);
end;

function FStr(Value: Double; Digits: Integer): string;
begin
  Str(Value:0:Digits, Result);
end;

function IZStr(Value, Width: Integer): string;
var
  sign: Boolean;
  p: PChar;
begin
  sign := Value < 0;
  if sign then
  begin
    Value := -Value;
    Dec(Width);
  end;
  Str(Value:Width, Result);
  p := PChar(Result);
  while p^ = ' ' do
  begin
    p^ := '0';
    Inc(p);
  end;
  if sign then
    Result := '-' + Result;
end;

function SInt(const S: string): Integer;
var
  code: Integer;
begin
  Val(S, Result, code);
end;

function SFloat(const S: string): Double;
var
  code: Integer;
begin
  Val(S, Result, code);
end;

// convert to 4 digits or 3 digits + point and unit character
function UnitFormat(Value: Comp): string;
const
  KiroSize = 1024;
  MegaSize = 1024 * 1024;
  GigaSize = 1024 * 1024 * 1024;
var
  u: Char;
  f: Integer;
  n: Double;
begin
  if Value < KiroSize then
    Str(Value:0:0, Result)
  else begin
    if Value >= MegaSize * 1000 then
    begin
      n := Value / GigaSize;
      u := 'G';
    end
    else if Value >= KiroSize * 1000 then
    begin
      n := Value / MegaSize;
      u := 'M';
    end
    else begin
      n := Value / KiroSize;
      u := 'K';
    end;

    // Truncate n because Str() round off
    if n >= 100 then
    begin
      f := 0;
      n := Int(n);
    end
    else if n >= 10 then
    begin
      f := 1;
      n := Int(n * 10) / 10;
    end
    else begin
      f := 2;
      n := Int(n * 100) / 100;
    end;
    Str(n:0:f, Result);
    Result := Result + u;
  end;
end;

function RoundOff(Value: Double): Longint;
begin
  if Value >= 0 then
    Result := Trunc(Value + 0.5)
  else
    Result := Trunc(Value - 0.5);
end;


{ String Function }

function ConvLineBreak(const S: string; const LineBreak: string): string;
var
  sp, ep, p: PChar;
  lblen, n: Integer;
begin
  lblen := Length(LineBreak);
  SetString(Result, nil, Length(S)+lblen-1);
  sp := @S[1];
  ep := @S[Length(S)+1];
  p := sp;
  n := 1;
  while p < ep do
  begin
    if p^ in [#10, #13] then
    begin
      Move(sp^, Result[n], p-sp);
      Inc(n, p-sp);
      Move(LineBreak[1], Result[n], lblen);
      Inc(n, lblen);
      if (p^ = #13) and ((p+1)^ = #10) then
        Inc(p);
      Inc(p);
      SetLength(Result, Length(Result)+lblen-1);
      sp := p;
    end else
      Inc(p);
  end;
  Move(sp^, Result[n], p-sp);
  Inc(n, p-sp);
  SetLength(Result, n-1);
end;

function CopyCStr(const Src: string; Dst: PChar; Size: Cardinal): Cardinal;
begin
  Result := Length(Src);
  if Result > (Size-1) then
    Result := Size-1;
  Move(PChar(Src)^, Dst^, Result);
  Dst[Result] := #0;
  Inc(Result);
end;


{ Path Function }

procedure SplitDrive(const Path: string; var Drive, Tail: string);
begin
  if (Length(Path) > 1) and (Path[2] = DriveSep) then
  begin
    Drive := Copy(Path, 1, 2);
    Tail := Copy(Path, 3, Length(Path)-2);
  end else
  begin
    Drive := '';
    Tail := Path;
  end;
end;

procedure SplitPath(const Path: string; var Head, Tail: string);
var
  i, dpos: Integer;
  p: string;
begin
  SplitDrive(Path, Head, p);
  dpos := 0;
  i := 0;
  while i < Length(p) do
  begin
    Inc(i);
    if p[i] in LeadBytes then
      Inc(i)
    else if p[i] = PathSep then
      dpos := i;
  end;
  Head := Head + Copy(p, 1, dpos-1);
  if (dpos = 1) or (dpos = 2) then  // C:\hoge or \\hoge
    Head := Head + PathSep;
  Tail := Copy(p, dpos+1, Length(p)-dpos);
end;

function BaseName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, s, Result);
end;

function DirName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, Result, s);
end;

procedure SplitExt(const Path: string; var Root, Ext: string);
var
  i: Integer;
  head, tail: string;
begin
  SplitPath(Path, head, tail);
  for i := Length(tail) downto 1 do
    if tail[i] = ExtSep then
    begin
      if (head <> '') and (RStr(head, 1) <> PathSep) then
        head := head + PathSep;
      Root := head + Copy(tail, 1, i-1);
      Ext := Copy(tail, i, Length(tail)-i+1);
      Exit;
    end;
  Root := Path;
  Ext := '';
end;

function JoinPath(const Path1: string; const Path2: string): string;
begin
  if RStr(Path1, 1) = PathSep then
    Result := Path1 + Path2
  else
    Result := Path1 + PathSep + Path2;
end;


{ ESException }

constructor ESException.Create(const AMsg: string);
begin
  Msg := AMsg;
end;


{ ESIOError }

constructor  ESIOError.Create(const AFileName: string);
begin
  ErrNo := SysErrNo;
  Msg := SysError(ErrNo);
  FileName := AFileName;
end;


{ ESOSError }

constructor  ESOSError.Create;
begin
  ErrNo := SysErrNo;
  Msg := SysError(ErrNo);
end;


{ TSFile }

destructor TSFile.Destroy;
begin
  if FBuf <> nil then
    FreeMem(FBuf);
  inherited;
end;

function TSFile.ReadLine(var S: string): Boolean;
const
  bufsize = 16384;
var
  len, dlen: Integer;
  sp: PChar;
  foundlf: Boolean;
begin
  if FBuf = nil then
    GetMem(FBuf, bufsize);

  dlen := 0;
  foundlf := False;
  while not foundlf do
  begin
    if FReadP = FEndP then
    begin
      len := Read(FBuf^, bufsize);
      if len = 0 then
      begin
        Result := dlen <> 0;
        Exit;
      end;
      FReadP := FBuf;
      FEndP := FBuf + len;
    end;

    sp := FReadP;
    while (FReadP < FEndP) and (not foundlf) do
    begin
      foundlf := FReadP^ = #10;
      Inc(FReadP);
    end;

    SetLength(S, dlen + (FReadP - sp));
    Move(sp^, S[dlen+1], FReadP - sp);
    Inc(dlen, FReadP - sp);
  end;
  Result := True;
end;

initialization
  InitLocale;

end.
