unit ifdib;

{ Support pixel formats:
    RGB 1,2,4,8,16(RGB555),24,32bit
    BITFIELDS 16, 32(only RGB_8888)bit
    RLE4, RLE8
}

interface

uses Windows, ImgFilter;

type
  PMaxBitmapInfo = ^TMaxBitmapInfo;
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    case Boolean of
      True: (MaskR, MaskG, MaskB: DWORD);
      False: (bmiColors: array[0..255+3] of TRGBQuad);
  end;

  TDibFilter = class(TImgFilter)
    constructor Create; override;
    function Test(Buf: PChar; BufSize: Cardinal): Boolean; override;
    procedure GetSize(Buf: PChar; BufSize: Cardinal;
      var Width: Cardinal; var Height: Cardinal;
      var Components: Cardinal); override;
    procedure Decode(InBuf: PChar; InBufSize: Cardinal; OutBuf: PChar;
      OutBufSize: Cardinal; var ImgInfo: TImgInfo); override;
  end;

function DecodeDIB(InBits: Pchar; InBitsSize: Cardinal;
  Info: TMaxBitmapInfo; OutBuf: Pchar; OutBufSize: Cardinal): Boolean;

implementation

type
  PWord = ^Word;
  PTriByte = ^TTriByte;
  TTriByte = array[0..2] of Byte;

function IsOS2Dib(Buf: PChar): Boolean;
begin
  with PBitmapInfo(@Buf[SizeOf(TBitmapFileHeader)])^ do
    Result := bmiHeader.biSize = SizeOf(TBitmapCoreHeader);
end;

procedure CoreInfoToBitmapInfo(BmInfo: PBitmapInfo; BmCore: PBitmapCoreInfo);
var
  i: Integer;
begin
  with BmInfo^ do
  begin
    FillChar(bmiHeader, SizeOf(bmiHeader), 0);
    with bmiHeader do
    begin
      biSize := SizeOf(bmiHeader);
      biWidth := BmCore.bmciHeader.bcWidth;
      biHeight := BmCore.bmciHeader.bcHeight;
      biPlanes := BmCore.bmciHeader.bcPlanes;
      biBitCount := BmCore.bmciHeader.bcBitCount;
      biCompression := BI_RGB;
    end;
    if (bmiHeader.biBitCount = 1) or (bmiHeader.biBitCount = 4)
       or (bmiHeader.biBitCount = 8) then
    begin
      for i := 0 to 1 shl bmiHeader.biBitCount - 1 do
      begin
        bmiColors[i].rgbBlue := BmCore.bmciColors[i].rgbtBlue;
        bmiColors[i].rgbGreen := BmCore.bmciColors[i].rgbtGreen;
        bmiColors[i].rgbRed := BmCore.bmciColors[i].rgbtRed;
      end;
    end;
  end;
end;

procedure GetBitmapInfo(Buf: PChar; var BitmapInfo: TMaxBitmapInfo);
begin
  if IsOS2Dib(Buf) then
    CoreInfoToBitmapInfo(@BitmapInfo, PBitmapCoreInfo(@Buf[SizeOf(TBitmapFileHeader)]))
  else
    Move(Buf[SizeOf(TBitmapFileHeader)], BitmapInfo, SizeOf(BitmapInfo));
end;

function CalcPitch(Width: Integer; Bpp: Integer): Integer;
begin
  Result := ((Width * Bpp + 7) shr 3 + 3) and not 3;
end;

procedure DecodeCLUT1(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x, m: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf >= EndBuf then
      Exit;
    m := 7 - x and 7; // place of bit
    Line^ := PTriByte(@Info.bmiColors[Byte(Buf^) shr m and 1])^;
    if m = 0 then
      Inc(Buf);
    Inc(Line);
  end;
end;

procedure DecodeCLUT4(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf >= EndBuf then
      Exit;
    Line^ := PTriByte(@Info.bmiColors[Byte(Buf^) shr ((1-x and 1) * 4) and $F])^;
    Inc(Line);
    Inc(Buf, x and 1);
  end;
end;

procedure DecodeCLUT8(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf >= EndBuf then
      Exit;
    Line^ := PTriByte(@Info.bmiColors[Byte(Buf^)])^;
    Inc(Line);
    Inc(Buf);
  end;
end;

procedure DecodeRGB888(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf > EndBuf-3 then
      Exit;
    Line^ := PTriByte(Buf)^;
    Inc(Line);
    Inc(Buf, 3);
  end;
end;

procedure DecodeRGB8888(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf > EndBuf-4 then
      Exit;
    Line^ := PTriByte(@Buf[1])^;
    Inc(Line);
    Inc(Buf, 4);
  end;
end;

function CountBits(Value: Cardinal): Integer;
begin
  Result := 0;
  repeat
    Inc(Result, Value and 1);
    Value := Value shr 1
  until Value = 0;
end;

function GetShift(Value: Cardinal): Integer;
begin
  Result := 0;
  Value := not Value;
  repeat
    Inc(Result, Value and 1);
    Value := Value shr 1
  until Value and 1 = 0;
end;

procedure DecodeBF16(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
  sp: PWord;
  rb, rs, gb, gs, bb, bs: Integer;
begin
  with Info do
  begin
    rb := CountBits(MaskR);
    rs := GetShift(MaskR);
    gb := CountBits(MaskG);
    gs := GetShift(MaskG);
    bb := CountBits(MaskB);
    bs := GetShift(MaskB);
    sp := PWord(Buf);
    //  43210(5bit) -> 43210432(8bit)
    // 543210(6bit) -> 54321054(8bit)
    for x := 0 to bmiHeader.biWidth-1 do
    begin
      if sp > EndBuf-2 then
        Exit;
      Line[0] := (sp^ and MaskB shr bs shl (8-bb)) or (sp^ and MaskB shr (bs+bb-(8-bb)));
      Line[1] := (sp^ and MaskG shr gs shl (8-gb)) or (sp^ and MaskG shr (gs+gb-(8-gb)));
      Line[2] := (sp^ and MaskR shr rs shl (8-rb)) or (sp^ and MaskR shr (rs+rb-(8-rb)));
      Inc(Line);
      Inc(sp);
    end;
  end;
end;

procedure DecodeRLE(InBuf: PChar; InBufSize: Integer; const Info: TMaxBitmapInfo;
  OutBuf: PChar; OutBufSize: Integer);
var
  dc: HDC;
  wnd: HWND;
  bmp: HBITMAP;
  bmi: TMaxBitmapInfo;
begin
  bmi := Info;

  // fix biSizeImage
  if bmi.bmiHeader.biSizeImage = 0 then
    bmi.bmiHeader.biSizeImage := InBufSize;

  wnd := GetDesktopWindow;
  dc := GetDC(wnd);
  bmp := CreateDIBitmap(dc, bmi.bmiHeader, CBM_INIT, InBuf, PBitmapInfo(@bmi)^,
    DIB_RGB_COLORS);
  try
    with bmi.bmiHeader do
    begin
      biBitCount := 24;
      biCompression := BI_RGB;
      biSizeImage := 0;
    end;
    GetDIBits(dc, bmp, 0, bmi.bmiHeader.biHeight, OutBuf, PBitmapInfo(@bmi)^,
      DIB_RGB_COLORS);
  finally
    DeleteObject(bmp);
    ReleaseDC(wnd, dc);
  end;
end;

function GetFormatName(Buf: PChar; BufSize: Longint): string;
var
  bmi: PMaxBitmapInfo;
  bci: PBitmapCoreInfo;
  s: string;
begin
  if IsOS2Dib(Buf) then
  begin
    bci := @Buf[SizeOf(TBitmapFileHeader)];
    Str(bci.bmciHeader.bcBitCount, s);
    Result := 'OS/2 Bitmap, ' + s + 'bit';
  end else
  begin
    bmi := @Buf[SizeOf(TBitmapFileHeader)];
    case bmi.bmiHeader.biCompression of
      BI_RGB:
        if bmi.bmiHeader.biBitCount = 16 then
          s := 'RGB555'
        else
          s := 'RGB888';
      BI_RLE8: s := 'RLE8';
      BI_RLE4: s := 'RLE4';
      BI_BITFIELDS:
        begin
          Str(CountBits(bmi.MaskR)*100+CountBits(bmi.MaskG)*10+CountBits(bmi.MaskB), s);
          s := 'RGB' + s + '(BITFIELDS)';
        end;
    else
      s := 'unknown';
    end;
    if bmi.bmiHeader.biHeight < 0 then
      s := s + ', topdown';
    Result := 'Windows Bitmap, ' + s;
  end;
end;

function DecodeDIB(InBits: PChar; InBitsSize: Cardinal;
  Info: TMaxBitmapInfo; OutBuf: Pchar; OutBufSize: Cardinal): Boolean;
var
  top, btm, tmp, endp, dline: PChar;
  spitch, dpitch, y: Cardinal;
  topdown: Boolean;
  decodeproc: procedure (Buf: PChar; EndBuf: PChar;
    const Info: TMaxBitmapInfo; Line: PTriByte);
begin
  decodeproc := nil;

  with Info.bmiHeader do
  begin
    spitch := CalcPitch(biWidth, biBitCount);
    dpitch := CalcPitch(biWidth, 24);
    endp := @InBits[InBitsSize];
    dline := OutBuf;
    topdown := biHeight < 0;
    if topdown then
      biHeight := -biHeight;

    case biCompression of
      BI_RGB:
        case biBitCount of
          1: decodeproc := DecodeCLUT1;
          4: decodeproc := DecodeCLUT4;
          8: decodeproc := DecodeCLUT8;
          16: begin // RGB555
                Info.MaskR := $7C00;
                Info.MaskG := $03E0;
                Info.MaskB := $001F;
                decodeproc := DecodeBF16;
              end;
          24: decodeproc := DecodeRGB888;
          32: decodeproc := DecodeRGB8888;
        end;
      BI_BITFIELDS:
        case biBitCount of
          16: decodeproc := DecodeBF16;
          32: if (Info.MaskR = $FF000000) and (Info.MaskG = $FF0000) and
                (Info.MaskB = $FF00) then
                decodeproc := DecodeRGB8888
        end;
    end;

    if (biCompression = BI_RLE4) or (biCompression = BI_RLE8) then
      DecodeRLE(inbits, InBitsSize, Info, OutBuf, OutBufSize)
    else if Assigned(decodeproc) then
      for y := 0 to biHeight-1 do
      begin
        decodeproc(InBits, endp, Info, PTriByte(dline));
        Inc(InBits, spitch);
        Inc(dline, dpitch);
      end
    else begin
      Result := False;
      Exit;
    end;

    // convert top down bmp
    if topdown then
    begin
      top := OutBuf;
      btm := OutBuf + (OutBufSize div dpitch - 1) * dpitch;
      GetMem(tmp, dpitch);
      while top < btm do
      begin
        Move(top^, tmp^, dpitch);
        Move(btm^, top^, dpitch);
        Move(tmp^, btm^, dpitch);
        Inc(top, dpitch);
        Dec(btm, dpitch);
      end;
      FreeMem(tmp);
    end;
  end;

  Result := True;
end;


{ TDibFilter }

constructor TDibFilter.Create;
begin
  FName := '(Internal DIB to DIB filter)';
end;

function TDibFilter.Test(Buf: PChar; BufSize: Cardinal): Boolean;
begin
  if BufSize < (SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader)) then
    Result := False
  else if (Buf[0] = 'B') and (Buf[1] = 'M') then
  begin
    Result := IsOS2Dib(Buf)
      or (PBitmapInfoHeader(@Buf[SizeOf(TBitmapFileHeader)]).biSize = SizeOf(TBitmapInfoHeader))
      or (PBitmapInfoHeader(@Buf[SizeOf(TBitmapFileHeader)]).biSize = SizeOf(TBitmapInfoHeader)+16);
      // biSize of BITFIELDS BMP which Photoshop Elements output is 56. (?)
  end else
    Result := False;
end;

procedure TDibFilter.GetSize(Buf: PChar; BufSize: Cardinal;
  var Width: Cardinal; var Height: Cardinal;
  var Components: Cardinal);
var
  bmi: TMaxBitmapInfo;
begin
  GetBitmapInfo(Buf, bmi);
  Width := bmi.bmiHeader.biWidth;
  if bmi.bmiHeader.biHeight < 0 then
    Height := -bmi.bmiHeader.biHeight
  else
    Height := bmi.bmiHeader.biHeight;
  Components := 3;
end;

procedure TDibFilter.Decode(InBuf: PChar; InBufSize: Cardinal; OutBuf: PChar;
  OutBufSize: Cardinal; var ImgInfo: TImgInfo);
var
  bmi: TMaxBitmapInfo;
begin
  GetBitmapInfo(InBuf, bmi);
  if not DecodeDIB(@InBuf[PBitmapFileHeader(InBuf).bfOffBits],
    InBufSize - PBitmapFileHeader(InBuf).bfOffBits, bmi, OutBuf, OutBufSize)
  then
    raise ESImgFilterError.Create('Invalid bitfields');

  ImgInfo.FormatName := GetFormatName(InBuf, InBufSize);
  ImgInfo.Bpp := bmi.bmiHeader.biBitCount;
  ImgInfo.Comment := '';
end;

end.
